//*****************************************************************************/
// Copyright (c) 1997-2005 Discreet Logic, Inc.
//
// These coded instructions, statements, and computer programs contain
// unpublished proprietary information written by Discreet Logic and
// are protected by Federal copyright law. They may not be disclosed
// to third parties or copied or duplicated in any form, in whole or
// in part, without the prior written consent of Discreet Logic.
//*****************************************************************************/

//==============================================================================
// EXTERNAL DECLARATIONS
//==============================================================================

#include "gfxUtils.h"
#include "fastmemcpy.h"
#include "convert.h"

#include <base/profiling/profilingInit.h>
#include <base/profiling/todTimer.h>
#include <base/profiling/hwIntervalTimer.h>
#include <base/profiling/prProcessTimer.h>
#include <base/profiling/hwProfiler.h>

#include <base/osal/osalInit.h>

#include <base/smartPtr/arrayPtr.h>
#include <base/smartPtr/objPtr.h>

#include "base/osal/getopt.h"

#include <iostream>
#include <assert.h>

#include <stdlib.h>
#include <string.h>

#include <GL/GLee.h>
#ifdef GLUT_FRAMEWORK
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

//==============================================================================
//	MACROS
//==============================================================================

#define BUFFER_OFFSET(i) ((char *)NULL + (i))

namespace {

DL_USING_NAMESPACE

//==============================================================================
// LOCAL DECLARATIONS
//==============================================================================

HwProfiler*  pMapProf(0x0);

/*----- classes -----*/

class TransferCmd {
public:

   virtual ~TransferCmd() {}

   virtual void init() = 0;
   virtual void execute() = 0;

   virtual GLenum getExtType() { return 0; };
   virtual GLenum getExtFmt() { return 0; };
   virtual GLenum getIntFmt() { return 0; };
   virtual GLenum getTexTarget() { return 0; };

   virtual std::string getUserDataTitle() { return "User Data"; };
   virtual std::string getUserData() { return ""; };
};

class TransferUsingPBOCmd  {
public:

   TransferUsingPBOCmd();
   virtual ~TransferUsingPBOCmd();

   void initPBO(unsigned char *img, int imgSize, GLenum target, GLenum usage);
   void copyToPBO(unsigned char* img, int imgSize);
   void bindPBO(GLenum target);
   void unbindPBO(GLenum target);
   void* mapPBO(GLenum target, GLuint access);
   bool unmapPBO(GLenum target);

protected:

   /*----- data members -----*/

   GLuint _pbo[16];
};

class MemoryCopyCmd : public TransferCmd {
public:

   enum Strategy { memcpy = 0, sse_memcpy = 1 } ;
   typedef void* (*bcopyFPtr)(void*, const void*, size_t);

   /*----- member functions -----*/

   MemoryCopyCmd(GLenum extFmt, GLenum pixelType, int width, int height,
                 unsigned int alignment, int strategy);

   virtual ~MemoryCopyCmd();

   virtual void init();
   virtual void execute();

   virtual GLenum getExtType();
   virtual GLenum getExtFmt();

   virtual std::string getUserDataTitle() { return "Strategy"; };
   virtual std::string getUserData();

protected:

   /*----- data members -----*/

   GLenum            _extFmt;
   GLenum            _pixelType;
   bcopyFPtr         _bcopy;
   AlignedBufferRing _srcBufs;
   AlignedBufferRing _destBufs;
   int               _imgSize;
   Strategy          _strategy;
};

class ConversionCmd : public TransferCmd {
public:

   enum Strategy {
       RGBAtoRGBA      = 0,
       RGBtoRGBA       = 1,
       intr_RGBAtoRGBA = 2,
       intr_RGBtoRGBA  = 3
   };
   typedef void (*conversionFuncPtr)(const unsigned char*, unsigned char*, int);

   /*----- member functions -----*/

   ConversionCmd(GLenum extFmt, GLenum pixelType, int width, int height,
                 unsigned int alignment, int strategy);

   virtual ~ConversionCmd();

   virtual void init();
   virtual void execute();

   virtual GLenum getExtType();
   virtual GLenum getExtFmt();

   virtual std::string getUserDataTitle() { return "Strategy"; };
   virtual std::string getUserData();

protected:

   /*----- data members -----*/

   GLenum            _extFmt;
   GLenum            _pixelType;
   conversionFuncPtr _conversionFunc;
   AlignedBufferRing _srcBufs;
   AlignedBufferRing _destBufs;
   int               _srcImgSize;
   int               _nbPixels;
   Strategy	         _strategy;
};

class MemoryCopyToPBOCmd : public MemoryCopyCmd, public TransferUsingPBOCmd {
public:

   /*----- member functions -----*/

   MemoryCopyToPBOCmd(GLenum extFmt, GLenum pixelType, int width, int height,
                      unsigned int alignment, int strategy);

   virtual ~MemoryCopyToPBOCmd();

   virtual void init();
   virtual void execute();
};

class ConversionToPBOCmd : public ConversionCmd, public TransferUsingPBOCmd {
public:

   /*----- member functions -----*/

   ConversionToPBOCmd(GLenum extFmt, GLenum pixelType, int width, int height,
                      unsigned int alignment, int strategy);

   virtual ~ConversionToPBOCmd();

   virtual void init();
   virtual void execute();
};

class DrawCmd : public TransferCmd {
public:

   /*----- member functions -----*/

  DrawCmd(GLenum extFmt, GLenum pixelType, int width, int height, 
          unsigned int alignment);

   virtual ~DrawCmd();

   virtual void init();
   virtual void execute();

   virtual GLenum getExtType();
   virtual GLenum getExtFmt();

protected:

   /*----- data members -----*/

   GLenum _extFmt;
   GLenum _pixelType;
   int    _width;
   int    _height;
   AlignedBufferRing _pixelBufs;
   int    _imgSize;
};

class DrawFromPBOCmd : public DrawCmd, public TransferUsingPBOCmd {
public:

   /*----- member functions -----*/

  DrawFromPBOCmd(GLenum extFmt, GLenum pixelType, int width, int height, 
                 unsigned int alignment);

   virtual ~DrawFromPBOCmd();

   virtual void init();
   virtual void execute();
};

class DrawWithPBOCmd : public DrawCmd, public TransferUsingPBOCmd {
public:

   /*----- member functions -----*/

  DrawWithPBOCmd(GLenum extFmt, GLenum pixelType, int width, int height, 
                 unsigned int alignment);

   virtual ~DrawWithPBOCmd();

   virtual void init();
   virtual void execute();
};

class DrawWithPBOConvertCmd : public DrawWithPBOCmd {
public:

   /*----- member functions -----*/

  DrawWithPBOConvertCmd(GLenum extFmt, GLenum pixelType, int width, int height, 
                        unsigned int alignment);

   virtual ~DrawWithPBOConvertCmd();

   virtual void init();
   virtual void execute();

   virtual bool isSupported() const;
};

class DrawWithPBOConvertExpandCmd : public DrawWithPBOCmd {
public:

   /*----- member functions -----*/

  DrawWithPBOConvertExpandCmd(GLenum extFmt, GLenum pixelType, int width, int height, 
                              unsigned int alignment);

   virtual ~DrawWithPBOConvertExpandCmd();

   virtual void init();
   virtual void execute();

   virtual bool isSupported() const;

protected:
   void fillPBO();

   /*----- data members -----*/

   GLenum _convExtFmt;
};

class DrawWithPBOConvertExpandTexCmd : public DrawWithPBOConvertExpandCmd {
public:

   /*----- member functions -----*/

  DrawWithPBOConvertExpandTexCmd(GLenum intFmt, GLenum extFmt, GLenum pixelType,
                                 int width, int height, 
                                 unsigned int alignment);

   virtual ~DrawWithPBOConvertExpandTexCmd();

   virtual void init();
   virtual void execute();

protected:

   /*----- data members -----*/

   GLuint _floatTexFP;
   GLenum _intFmt;
};

class DrawWithPBOConvertExpandSubTexCmd : public DrawWithPBOConvertExpandTexCmd {
public:

   /*----- member functions -----*/

  DrawWithPBOConvertExpandSubTexCmd(GLenum iFmt, GLenum eFmt, GLenum pixelType,
                                    int width, int height, 
                                    unsigned int alignment);

   virtual ~DrawWithPBOConvertExpandSubTexCmd();

   virtual void execute();

protected:

   /*----- data members -----*/

   GLuint _tex;
};

// Abstract base class for all texture load commands.

class TexLoadCmdBase : public TransferCmd {
public:

   TexLoadCmdBase(
      GLenum intFmt,
      GLenum extFmt,
      GLenum pixelType,
      int width,
      int height,
      unsigned int alignment
   );

   // Abstract class, can't create instance of it.
   virtual ~TexLoadCmdBase() = 0;

   virtual GLenum getExtType();
   virtual GLenum getExtFmt();
   virtual GLenum getIntFmt();
   virtual GLenum getTexTarget();

   void bindFloatTexFP();
   void unbindFloatTexFP();

protected:

   /*----- data members -----*/

   GLenum _texTarget;
   GLenum _intFmt;
   GLenum _extFmt;
   int    _width;
   int    _height;
   AlignedBufferRing _pixelBufs;
   int    _imgSize;
   GLenum _pixelType;
   GLuint _floatTexFP;
};

class TexLoadCmd : public TexLoadCmdBase {
public:

   /*----- member functions -----*/

   TexLoadCmd(
      GLenum intFmt,
      GLenum extFmt,
      GLenum pixelType,
      int width,
      int height,
      unsigned int alignment
   );

   virtual ~TexLoadCmd();

   virtual void init();
   virtual void execute();
};

class TexLoadFromPBOCmd : public TexLoadCmd, public TransferUsingPBOCmd {
public:

   /*----- member functions -----*/

   TexLoadFromPBOCmd(
      GLenum intFmt,
      GLenum extFmt,
      GLenum pixelType,
      int width,
      int height,
      unsigned int alignment
   );

   virtual ~TexLoadFromPBOCmd();

   virtual void init();
   virtual void execute();
};

class TexLoadWithPBOCmd : public TexLoadCmd, public TransferUsingPBOCmd {
public:

   /*----- member functions -----*/

   TexLoadWithPBOCmd(
      GLenum intFmt,
      GLenum extFmt,
      GLenum pixelType,
      int width,
      int height,
      unsigned int alignment
   );

   virtual ~TexLoadWithPBOCmd();

   virtual void init();
   virtual void execute();
};

class TexLoadUseCmd : public TexLoadCmd {
public:

   /*----- member functions -----*/

   TexLoadUseCmd(
      GLenum intFmt,
      GLenum extFmt,
      GLenum pixelType,
      int width,
      int height,
      unsigned int alignment
   );

   virtual ~TexLoadUseCmd();

   virtual void init();
   virtual void execute();

protected:

   /*----- data members -----*/

   int _x;
   int _y;
};

class TexLoadUseFromPBOCmd : public TexLoadUseCmd, public TransferUsingPBOCmd {
public:

   /*----- member functions -----*/

   TexLoadUseFromPBOCmd(
      GLenum intFmt,
      GLenum extFmt,
      GLenum pixelType,
      int width,
      int height,
      unsigned int alignment
   );

   virtual ~TexLoadUseFromPBOCmd();

   virtual void init();
   virtual void execute();
};

class TexLoadUseWithPBOCmd : public TexLoadUseCmd, public TransferUsingPBOCmd {
public:

   /*----- member functions -----*/

   TexLoadUseWithPBOCmd(
      GLenum intFmt,
      GLenum extFmt,
      GLenum pixelType,
      int width,
      int height,
      unsigned int alignment
   );

   virtual ~TexLoadUseWithPBOCmd();

   virtual void init();
   virtual void execute();
};

class TexLoadUseWithPBOConvertCmd : public TexLoadUseWithPBOCmd {
public:

   /*----- member functions -----*/

   TexLoadUseWithPBOConvertCmd(
      GLenum intFmt,
      GLenum extFmt,
      GLenum pixelType,
      int width,
      int height,
      unsigned int alignment
   );

   virtual ~TexLoadUseWithPBOConvertCmd();

   virtual void init();
   virtual void execute();

   virtual bool isSupported() const;

};

class TexLoadUseWithPBOConvertExpandCmd : public TexLoadUseWithPBOCmd {
public:

   /*----- member functions -----*/

   TexLoadUseWithPBOConvertExpandCmd(
      GLenum intFmt,
      GLenum extFmt,
      GLenum pixelType,
      int width,
      int height,
      unsigned int alignment
   );

   virtual ~TexLoadUseWithPBOConvertExpandCmd();

   virtual void init();
   virtual void execute();

   virtual bool isSupported() const;

protected:

   /*----- data members -----*/

   int    _nbPixels;
   GLenum _convExtFmt;
};

class FillCmd : public TransferCmd {
public:

   /*----- member functions -----*/

   FillCmd(bool pixelAlpha, int width, int height);

   virtual ~FillCmd();

   virtual void init();
   virtual void execute();

   virtual GLenum getExtFmt();

private:

   /*----- data members -----*/

   int  _width;
   int  _height;
   bool _alpha;
};

class TexFillCmd : public TransferCmd {
public:

   /*----- member functions -----*/

   TexFillCmd(GLenum intFmt,
              GLenum extFmt,
              GLenum pixelType,
              int width,
              int height,
              unsigned int alignment);

   virtual ~TexFillCmd();

   virtual void init();
   virtual void execute();

   virtual GLenum getIntFmt();
   virtual GLenum getExtFmt();
   virtual GLenum getExtType();
   virtual GLenum getTexTarget();

protected:

   /*----- data members -----*/

   GLenum _texTarget;
   GLenum _intFmt;
   GLenum _extFmt;
   int    _width;
   int    _height;
   float  _x0;
   float  _y0;
   AlignedBufferRing _pixelBufs;
   GLenum _pixelType;
};

class ReadCmd : public TransferCmd {
public:

   /*----- member functions -----*/

   ReadCmd(GLenum extFmt, GLenum pixelType, int width, int height, 
           unsigned int alignment);

   virtual ~ReadCmd();

   virtual void init();
   virtual void execute();

   virtual GLenum getExtType();
   virtual GLenum getExtFmt();

private:

   /*----- data members -----*/

   GLenum _extFmt;
   int _width;
   int _height;
   AlignedBufferRing _pixelBufs;
   GLenum _pixelType;
};

/*----- function prototypes -----*/

void printUsage(const char* argv0);

// Create a graphics command object to benchmark, given a command-line
// character option.
ObjPtrRet<TransferCmd> cmdFactory(
   int option,
   GLenum intFmt,
   GLenum extFmt,
   GLenum pixelType,
   int width,
   int height,
   unsigned int alignment
);

//==============================================================================
// LOCAL VARIABLES
//==============================================================================

ObjPtr<TransferCmd> *gfxCmd = 0;
GLenum extFormat =  GL_RGBA;
GLenum intFormat = GL_RGBA8;
int nbIterations = 0;
int windowId = -1;
int width = 512;
int height = 512;
float zoom = 1.0f;
unsigned int nbPBOs = ( getenv( "DL_USE_SINGLE_PBO" ) != 0 ? 1 : 16 );
unsigned int nbSrcBufs = 16;
unsigned int nbDestBufs = 16;
bool glFinishBetweenIterations = getenv( "DL_GLFINISH_ITERS" )!=0;

//==============================================================================
// LOCAL FUNCTIONS
//==============================================================================

//------------------------------------------------------------------------------
//
void printUsage(const char* argv0)
{
   std::cerr << "usage:\n" << argv0
             << " [-v] [-x width] [-y height] [-z zoom] [-e extFmt] [-i intFmt] [-p type]"
             << " [-a alignment] [-u userdata] [-g] [-S] "
             << "{-c | -d | -f | -h | -j | -k | -l | -m | -n | -r | -t | -w |"
             << " -b | -o | -s | -q | -A | -B | -C | -D | -E | -F | -G | -H} "
             << "iterations\n\n"
             << "where:\n"
             << "-c\t\t Memory copy test\n"
             << "-d\t\t Memory copy to PBO test\n"
             << "-F\t\t Conversion test\n"
             << "-G\t\t Conversion to PBO test\n"
             << "-f\t\t Fill rate test\n"
             << "-h\t\t Texture load/use test\n"
             << "-j\t\t Texture load/use from PBO test\n"
             << "-k\t\t Texture load/use with PBO test\n"
             << "-l\t\t Texture load test\n"
             << "-m\t\t Texture load from PBO test\n"
             << "-n\t\t Texture load with PBO test\n"
             << "-C\t\t Texture load with PBO and conversion from u16 to HF test\n"
             << "-D\t\t Texture load with PBO and conversion from RGB u16 to RGBA HF test\n"
             << "-r\t\t Read test\n"
             << "-t\t\t Textured fill test\n"
             << "-w\t\t Write test\n"
             << "-b\t\t Write from PBO test\n"
             << "-o\t\t Write with PBO test\n"
             << "-s\t\t Write test with MultiSampling fastest\n"
             << "-q\t\t Write test with MultiSampling nicest\n"
             << "-A\t\t Write test with PBO and conversion from u16 to HF\n"
             << "-H\t\t Write test with PBO and conversion from RGB u16 to RGBA HF\n"
             << "-B\t\t Write test with PBO and conversion from RGB u16 to RGBA HF (use glTexImage2D to draw)\n"
             << "-E\t\t Write test with PBO and conversion from RGB u16 to RGBA HF (use glTexSubImage2D to draw)\n"
             << "-x width\t Width (default 512)\n"
             << "-y height\t Height (default 512)\n"
             << "-a alignment\t System memory buffer byte alignment "
             << "(default 8)\n"
             << "-u userdata\t Supply user data to test\n"
             << "-g\t\t force a call to glFinish between each iteration\n"
             << "-S\t\t force the use of the same single PBO for every iteration\n"
             << "-e extFmt\t Specify pixel/texel external format, for "
             << "pixel reads, writes,\n\t\t and texture loads (see "
             << "glTexImage2D --- default GL_RGBA), or\n\t\t whether to "
             << "fill with destination alpha, for fill tests (in\n\t\t which "
             << "case any enumerant with alpha says fill destination\n\t\t "
             << "alpha)\n"
             << "-i intFmt \t Specify texel internal format, for texture "
             << "loads (see\n\t\t glTexImage2D --- default GL_RGBA8)\n"
             << "-p type\t\t Specify pixel/texel external type, for "
             << "pixel reads, writes,\n\t\t texture loads (see "
             << "glTexImage2D --- default GL_UNSIGNED_BYTE)\n"
             << "-z zoom \t Specify pixel zoom value\n\n"

             << "Note how suboptions must precede the test option, e.g.\n\n"
             << argv0 << " -x 256 -y 256 -e GL_RGB -i GL_RGB8 -l 500\n";
}


//------------------------------------------------------------------------------
//
ObjPtrRet<TransferCmd> cmdFactory(
   int option,
   GLenum intFmt,
   GLenum extFmt,
   GLenum pixelType,
   int width,
   int height,
   unsigned int alignment,
   int userData
)
{
   ObjPtr<TransferCmd> cmd;

   switch (option) {
      case 'c': {
         cmd = ObjPtrRet<MemoryCopyCmd>(
            new MemoryCopyCmd(extFmt, pixelType, width, height, alignment, userData)
         );
      } break;

      case 'd': {
         cmd = ObjPtrRet<MemoryCopyToPBOCmd>(
            new MemoryCopyToPBOCmd(extFmt, pixelType, width, height, alignment, userData)
         );
      } break;

      case 'f': {
         cmd = ObjPtrRet<FillCmd>(
            new FillCmd(hasAlpha(extFmt), width, height)
         );
      } break;

      case 'h': {
         cmd = ObjPtrRet<TexLoadUseCmd>(
	    new TexLoadUseCmd(intFmt, extFmt, pixelType, width, height, alignment)
         );
      } break;

      case 'j': {
         cmd = ObjPtrRet<TexLoadUseFromPBOCmd>(
	    new TexLoadUseFromPBOCmd(intFmt, extFmt, pixelType,
                                     width, height, alignment)
         );
      } break;

      case 'k': {
         cmd = ObjPtrRet<TexLoadUseWithPBOCmd>(
	    new TexLoadUseWithPBOCmd(intFmt, extFmt, pixelType,
                                     width, height, alignment)
         );
      } break;

      case 'l': {
         cmd = ObjPtrRet<TexLoadCmd>(
	    new TexLoadCmd(intFmt, extFmt, pixelType, width, height, alignment)
         );
      } break;

      case 'm': {
         cmd = ObjPtrRet<TexLoadFromPBOCmd>(
	    new TexLoadFromPBOCmd(intFmt, extFmt, pixelType, width, height, alignment)
         );
      } break;

      case 'n': {
         cmd = ObjPtrRet<TexLoadWithPBOCmd>(
	    new TexLoadWithPBOCmd(intFmt, extFmt, pixelType, width, height, alignment)
         );
      } break;

      case 'r': {
         cmd = ObjPtrRet<ReadCmd>(
	    new ReadCmd(extFmt, pixelType, width, height, alignment)
         );
      } break;

      case 't': {
         cmd = ObjPtrRet<TexFillCmd>(
	    new TexFillCmd(intFmt, extFmt, pixelType, width, height, alignment)
         );
      } break;

      case 's':
      case 'q':
      case 'w': {
         cmd = ObjPtrRet<DrawCmd>(
	    new DrawCmd(extFmt, pixelType, width, height, alignment)
         );
      } break;

      case 'b': {
         cmd = ObjPtrRet<DrawFromPBOCmd>(
	    new DrawFromPBOCmd(extFmt, pixelType, width, height, alignment)
         );
      } break;

      case 'o': {
         cmd = ObjPtrRet<DrawWithPBOCmd>(
	    new DrawWithPBOCmd(extFmt, pixelType, width, height, alignment)
         );
      } break;

      case 'A': {
         cmd = ObjPtrRet<DrawWithPBOConvertCmd>(
	    new DrawWithPBOConvertCmd(extFmt, pixelType, width, height, alignment)
         );
      } break;

      case 'B': {
         cmd = ObjPtrRet<DrawWithPBOConvertExpandTexCmd>(
	    new DrawWithPBOConvertExpandTexCmd(intFmt, extFmt, pixelType,
                                               width, height, alignment)
         );
      } break;

      case 'C': {
         cmd = ObjPtrRet<TexLoadUseWithPBOConvertCmd>(
	    new TexLoadUseWithPBOConvertCmd(intFmt, extFmt, pixelType,
                                            width, height, alignment)
         );
      } break;

      case 'D': {
         cmd = ObjPtrRet<TexLoadUseWithPBOConvertExpandCmd>(
	    new TexLoadUseWithPBOConvertExpandCmd(intFmt, extFmt, pixelType,
                                                  width, height, alignment)
         );
      } break;

      case 'E': {
         cmd = ObjPtrRet<DrawWithPBOConvertExpandSubTexCmd>(
	    new DrawWithPBOConvertExpandSubTexCmd(intFmt, extFmt, pixelType,
                                                  width, height, alignment)
         );
      } break;

      case 'F': {
         cmd = ObjPtrRet<ConversionCmd>(
	    new ConversionCmd(extFmt, pixelType, width, height, alignment, userData)
         );
      } break;

      case 'G': {
         cmd = ObjPtrRet<ConversionCmd>(
	    new ConversionToPBOCmd(extFmt, pixelType, width, height, alignment, userData)
         );
      } break;

      case 'H': {
         cmd = ObjPtrRet<DrawWithPBOConvertExpandCmd>(
	    new DrawWithPBOConvertExpandCmd(extFmt, pixelType, width, height, alignment)
         );
      } break;

   }

   return ObjPtrRet<TransferCmd>(cmd);
}

//==============================================================================
//	CLASS TransferUsingPBOCmd
//==============================================================================

//------------------------------------------------------------------------------
//
TransferUsingPBOCmd::TransferUsingPBOCmd()
{
   assert( nbPBOs <= sizeof(_pbo)/sizeof(_pbo[0]) );
   glGenBuffersARB( nbPBOs, _pbo );
}

//------------------------------------------------------------------------------
//
TransferUsingPBOCmd::~TransferUsingPBOCmd()
{
   glDeleteBuffersARB( nbPBOs, _pbo );
}

//------------------------------------------------------------------------------
//
void TransferUsingPBOCmd::initPBO(unsigned char* img, int imgSize, GLenum target, GLenum usage)
{
   // We know that each bind will bind the next PBO in the PBO ring.  Repeat
   // for the entire size of the PBO ring so that we know each is initialized.
   for ( unsigned int i=0; i<nbPBOs; ++i ) {
      bindPBO(target);
      glBufferDataARB( target, imgSize, NULL, usage );
      if (img)
         copyToPBO(img, imgSize);

      unbindPBO(target);
   }
}

//------------------------------------------------------------------------------
//
void TransferUsingPBOCmd::copyToPBO(unsigned char* img, int imgSize)
{
   char *pPixelsPBO = static_cast<char *>(mapPBO(GL_PIXEL_UNPACK_BUFFER_EXT, GL_WRITE_ONLY));
   if (NULL == pPixelsPBO)
   {
      std::cerr << "[error] Couldn't map pixel buffer. Exiting\n";
      exit( EXIT_FAILURE );
   }

   ::sse_memcpy(pPixelsPBO, img, imgSize);

   if (!unmapPBO(GL_PIXEL_UNPACK_BUFFER_EXT))
   {
      std::cerr << "[error] Couldn't unmap pixel buffer. Exiting\n";
      exit( EXIT_FAILURE );
   }
}

//------------------------------------------------------------------------------
//
void TransferUsingPBOCmd::bindPBO(GLenum target)
{
   // Bind to a diferent PBO on every iteration.  This allows PBO use to be
   // pipelined, and also avoids the possibility of the CPU caching it.
   static unsigned int pboIndex = 0;
   pboIndex = ( pboIndex + 1 ) % nbPBOs;
   glBindBufferARB( target, _pbo[pboIndex] );
}

//------------------------------------------------------------------------------
//
void TransferUsingPBOCmd::unbindPBO(GLenum target)
{
   glBindBufferARB(target, 0);
}

//------------------------------------------------------------------------------
//
void* TransferUsingPBOCmd::mapPBO(GLenum target, GLuint access)
{
   pMapProf->start();
   void* p = glMapBufferARB(target, access);
   pMapProf->stop();
   pMapProf->endCycle();
   return p;
}

//------------------------------------------------------------------------------
//
bool TransferUsingPBOCmd::unmapPBO(GLenum target)
{
   return glUnmapBufferARB(target);
}

//==============================================================================
//	CLASS MemoryCopyCmd
//==============================================================================

//------------------------------------------------------------------------------
//
MemoryCopyCmd::MemoryCopyCmd(GLenum extFmt, GLenum pixelType, int width, int height,
                             unsigned int alignment, int strategy)
:
   TransferCmd(),
   _extFmt(extFmt),
   _pixelType(pixelType),
   _bcopy(NULL),
   _srcBufs( nbSrcBufs,
             (hasAlpha(extFmt) ? 4 : 3)*width*height*bytesPerChannel(pixelType),
             alignment ),
   _destBufs( nbDestBufs,
              (hasAlpha(extFmt) ? 4 : 3)*width*height*bytesPerChannel(pixelType),
              alignment ),
   _imgSize((hasAlpha(extFmt) ? 4 : 3)*width*height*bytesPerChannel(pixelType)),
   _strategy(Strategy(strategy))
{
   switch (_strategy)
   {
      case memcpy:
      default:
         _bcopy = ::memcpy;
         break;
      case sse_memcpy:
         _bcopy = ::sse_memcpy;
         break;
   }
}

//------------------------------------------------------------------------------
//
MemoryCopyCmd::~MemoryCopyCmd()
{
}

//------------------------------------------------------------------------------
//
void MemoryCopyCmd::init()
{
   // Memset all the buffers so that the OS does not keep them specially marked
   // as unitialized.  Some OS's try to be smart and optimize access and copying
   // from unitilized buffers (hence skewing our benchmark results)
   for ( unsigned int i=0; i<_srcBufs.size(); ++i ) {
      memset( _srcBufs.get(i), 0, _imgSize );
   }
}

//------------------------------------------------------------------------------
//
void MemoryCopyCmd::execute()
{
   // Use different buffers on each execution so that the CPU can't cache the
   // data.
   static unsigned int bufIndex = 0;
   bufIndex++;
   _bcopy( _destBufs.get(bufIndex), _srcBufs.get(bufIndex), _imgSize );
}

//------------------------------------------------------------------------------
//
GLenum MemoryCopyCmd::getExtType()
{
   return _pixelType;
}

//------------------------------------------------------------------------------
//
GLenum MemoryCopyCmd::getExtFmt()
{
   return _extFmt;
}

//------------------------------------------------------------------------------
//
std::string MemoryCopyCmd::getUserData()
{
   switch (_strategy)
   {
      case memcpy:
      default:
         return "memcpy";
         break;
      case sse_memcpy:
         return "sse_memcpy";
         break;
   }
}

//==============================================================================
//	CLASS ConversionCmd
//==============================================================================

//------------------------------------------------------------------------------
//
ConversionCmd::ConversionCmd(GLenum extFmt, GLenum pixelType, int width, int height,
                             unsigned int alignment, int strategy)
:
   TransferCmd(),
   _extFmt(extFmt),
   _pixelType(pixelType),
   _conversionFunc(NULL),
   _srcBufs( nbSrcBufs,
             (hasAlpha(extFmt) ? 4 : 3)*width*height*bytesPerChannel(pixelType),
             alignment ),
   _destBufs( nbDestBufs,
              (hasAlpha(extFmt) ? 4 : 3)*width*height*bytesPerChannel(pixelType),
              alignment ),
   _srcImgSize((hasAlpha(extFmt) ? 4 : 3)*width*height*bytesPerChannel(pixelType)),
   _nbPixels(width*height),
   _strategy(Strategy(strategy))
{
   if ( (_extFmt != GL_RGB && _extFmt != GL_RGBA) ||
        _pixelType != GL_UNSIGNED_SHORT           ||
        alignment < 16) {
      std::cerr << "[error] Invalid parameters for this test. Exiting\n";
      exit( EXIT_FAILURE );
   }

   switch (_strategy)
   {
   case RGBAtoRGBA:
   default:
      _conversionFunc = cv_sse_RGBA16toRGBAHF;
      // The conversion function "expects" an alpha channel
      if (_extFmt == GL_RGB)
         _nbPixels = _nbPixels * 3 / 4;
      break;
   case RGBtoRGBA:
      if ( _extFmt != GL_RGB ) {
         std::cerr << "[error] This test expects RGB input values. Exiting\n";
         exit( EXIT_FAILURE );
      }
      _conversionFunc = cv_sse_RGB16toRGBAHF;
      break;
   case intr_RGBAtoRGBA:
      _conversionFunc = cv_sse_intr_RGBA16toRGBAHF;
      // The conversion function "expects" an alpha channel
      if (_extFmt == GL_RGB)
         _nbPixels = _nbPixels * 3 / 4;
      break;
   case intr_RGBtoRGBA:
      if ( _extFmt != GL_RGB ) {
         std::cerr << "[error] This test expects RGB input values. Exiting\n";
         exit( EXIT_FAILURE );
      }
      _conversionFunc = cv_sse_intr_RGB16toRGBAHF;
      break;
   }
}

//------------------------------------------------------------------------------
//
ConversionCmd::~ConversionCmd()
{
}

//------------------------------------------------------------------------------
//
void ConversionCmd::init()
{
   // Memset all the buffers so that the OS does not keep them specially marked
   // as unitialized.  Some OS's try to be smart and optimize access and copying
   // from unitilized buffers (hence skewing our benchmark results)
   for ( unsigned int i=0; i<_srcBufs.size(); ++i ) {
      memset( _srcBufs.get(i), 0, _srcImgSize );
   }
}

//------------------------------------------------------------------------------
//
void ConversionCmd::execute()
{
   // Use a different buffer on each execution so that the CPU can't cache the
   // data.
   static unsigned int bufIndex = 0;
   bufIndex++;
   _conversionFunc( _srcBufs.get(bufIndex), _destBufs.get(bufIndex), _nbPixels );
}

//------------------------------------------------------------------------------
//
GLenum ConversionCmd::getExtType()
{
   return _pixelType;
}

//------------------------------------------------------------------------------
//
GLenum ConversionCmd::getExtFmt()
{
   return _extFmt;
}

//------------------------------------------------------------------------------
//
std::string ConversionCmd::getUserData()
{
   switch (_strategy)
   {
   case RGBAtoRGBA:
   default:
      return "RGBA to RGBA";
   case RGBtoRGBA:
      return "RGB to RGBA";
   case intr_RGBAtoRGBA:
      return "intr RGBA to RGBA";
   case intr_RGBtoRGBA:
      return "intr RGB to RGBA";
   }
}

//==============================================================================
//	CLASS MemoryCopyToPBOCmd
//==============================================================================

//------------------------------------------------------------------------------
//
MemoryCopyToPBOCmd::MemoryCopyToPBOCmd(GLenum extFmt, GLenum pixelType, int width, int height,
                      unsigned int alignment, int strategy)
:
   MemoryCopyCmd(extFmt, pixelType, width, height, alignment, strategy),
   TransferUsingPBOCmd()
{
}

//------------------------------------------------------------------------------
//
MemoryCopyToPBOCmd::~MemoryCopyToPBOCmd()
{}

//------------------------------------------------------------------------------
//
void MemoryCopyToPBOCmd::init()
{
   MemoryCopyCmd::init();
   TransferUsingPBOCmd::initPBO( _srcBufs.get(0), _imgSize,
                                 GL_PIXEL_UNPACK_BUFFER_EXT,
                                 GL_DYNAMIC_DRAW_ARB );
}

//------------------------------------------------------------------------------
//
void MemoryCopyToPBOCmd::execute()
{
   TransferUsingPBOCmd::bindPBO(GL_PIXEL_UNPACK_BUFFER_EXT);

   char *pPixelsPBO = static_cast<char *>(TransferUsingPBOCmd::mapPBO(GL_PIXEL_UNPACK_BUFFER_EXT, GL_WRITE_ONLY));
   if (NULL == pPixelsPBO)
   {
      std::cerr << "[error] Couldn't map pixel buffer. Exiting\n";
      exit( EXIT_FAILURE );
   }

   // Use a different buffer on each execution so that the CPU can't cache the
   // data.
   static unsigned int srcIndex = 0;
   srcIndex++;
   _bcopy( pPixelsPBO, _srcBufs.get(srcIndex), _imgSize );

   if (!TransferUsingPBOCmd::unmapPBO(GL_PIXEL_UNPACK_BUFFER_EXT))
   {
      std::cerr << "[error] Couldn't unmap pixel buffer. Exiting\n";
      exit( EXIT_FAILURE );
   }

   TransferUsingPBOCmd::unbindPBO(GL_PIXEL_UNPACK_BUFFER_EXT);
}

//==============================================================================
//	CLASS ConversionToPBOCmd
//==============================================================================

//------------------------------------------------------------------------------
//
ConversionToPBOCmd::ConversionToPBOCmd(GLenum extFmt, GLenum pixelType, int width, int height,
                                       unsigned int alignment, int strategy)
:
   ConversionCmd(extFmt, pixelType, width, height, alignment, strategy),
   TransferUsingPBOCmd()
{
}

//------------------------------------------------------------------------------
//
ConversionToPBOCmd::~ConversionToPBOCmd()
{}

//------------------------------------------------------------------------------
//
void ConversionToPBOCmd::init()
{
   ConversionCmd::init();
   TransferUsingPBOCmd::initPBO( _srcBufs.get(0),
                                 _nbPixels*4*sizeof(unsigned short),
                                 GL_PIXEL_UNPACK_BUFFER_EXT,
                                 GL_DYNAMIC_DRAW_ARB );
}

//------------------------------------------------------------------------------
//
void ConversionToPBOCmd::execute()
{
   TransferUsingPBOCmd::bindPBO(GL_PIXEL_UNPACK_BUFFER_EXT);

   unsigned char *pPixelsPBO = static_cast<unsigned char *>(TransferUsingPBOCmd::mapPBO(GL_PIXEL_UNPACK_BUFFER_EXT, GL_WRITE_ONLY));
   if (NULL == pPixelsPBO)
   {
      std::cerr << "[error] Couldn't map pixel buffer. Exiting\n";
      exit( EXIT_FAILURE );
   }

   // Use a different buffer on each execution so that the CPU can't cache the
   // data.
   static unsigned int srcIndex = 0;
   srcIndex++;
   _conversionFunc( _srcBufs.get(srcIndex), pPixelsPBO, _nbPixels );

   if (!TransferUsingPBOCmd::unmapPBO(GL_PIXEL_UNPACK_BUFFER_EXT))
   {
      std::cerr << "[error] Couldn't unmap pixel buffer. Exiting\n";
      exit( EXIT_FAILURE );
   }

   TransferUsingPBOCmd::unbindPBO(GL_PIXEL_UNPACK_BUFFER_EXT);
}

//==============================================================================
// CLASS DrawCmd
//==============================================================================

//------------------------------------------------------------------------------
//
DrawCmd::DrawCmd(GLenum extFmt, GLenum pixelType, int width, int height, unsigned int alignment) :
  TransferCmd(), _extFmt(extFmt), _pixelType(pixelType), _width(width), 
  _height(height), 
  _pixelBufs( nbSrcBufs,
              (hasAlpha(extFmt) ? 4 : 3)*width*height*bytesPerChannel(pixelType),
              alignment ),
  _imgSize((hasAlpha(extFmt) ? 4 : 3)*width*height*bytesPerChannel(pixelType))
{
  for ( unsigned int i=0; i<_pixelBufs.size(); ++i ) {
    switch(pixelType){
    case GL_UNSIGNED_BYTE:
    default:
      fillImage(_extFmt, _pixelBufs.get(i), _width, _height, Color3ub(220, 20, 60));
      break;
    case GL_UNSIGNED_SHORT:
      fillImage(_extFmt, _pixelBufs.get(i), _width, _height, 
                Color3us(56320, 5120, 15360));
      break;
    case GL_HALF_FLOAT_ARB:
      fillImageHalf(_extFmt, _pixelBufs.get(i), _width, _height, 
                    Color3f(0.859f, 0.078f, 0.234f));
      break;
    case GL_UNSIGNED_INT:
      fillImage(_extFmt, _pixelBufs.get(i), _width, _height, 
                Color3ui(3690000000U, 335000000U, 1000000000U));
      break;
#ifdef GL_EXT_packed_pixels
    case GL_UNSIGNED_INT_10_10_10_2_EXT:
      fillImage(_extFmt, _pixelBufs.get(i), _width, _height, 
                Color4_10bit(880, 80, 240, 1023));
      break;
#endif
    case GL_FLOAT:
      fillImage(_extFmt, _pixelBufs.get(i), _width, _height, 
                Color3f(0.859f, 0.078f, 0.234f));
      break;
    }
  }
}

//------------------------------------------------------------------------------
//
DrawCmd::~DrawCmd()
{
}

//------------------------------------------------------------------------------
//
void DrawCmd::init()
{
}

//------------------------------------------------------------------------------
//
void DrawCmd::execute()
{
   // Use a different buffer on each execution so that the CPU can't cache the
   // data.
   static unsigned int srcIndex = 0;
   srcIndex++;
   glDrawPixels(_width, _height, _extFmt, _pixelType, _pixelBufs.get(srcIndex));
}

//------------------------------------------------------------------------------
//
GLenum DrawCmd::getExtType()
{
   return _pixelType;
}

//------------------------------------------------------------------------------
//
GLenum DrawCmd::getExtFmt()
{
   return _extFmt;
}

//==============================================================================
// CLASS DrawFromPBOCmd
//==============================================================================

//------------------------------------------------------------------------------
//
DrawFromPBOCmd::DrawFromPBOCmd(GLenum extFmt, GLenum pixelType, int width,
                               int height, unsigned int alignment)
:
   DrawCmd(extFmt, pixelType, width, height, alignment),
   TransferUsingPBOCmd()
{}

//------------------------------------------------------------------------------
//
DrawFromPBOCmd::~DrawFromPBOCmd()
{}

//------------------------------------------------------------------------------
//
void DrawFromPBOCmd::init()
{
   DrawCmd::init();
   TransferUsingPBOCmd::initPBO(_pixelBufs.get(0), _imgSize, GL_PIXEL_UNPACK_BUFFER_EXT, GL_DYNAMIC_DRAW_ARB);
}

//------------------------------------------------------------------------------
//
void DrawFromPBOCmd::execute()
{
   TransferUsingPBOCmd::bindPBO(GL_PIXEL_UNPACK_BUFFER_EXT);

   glDrawPixels(_width, _height, _extFmt, _pixelType, BUFFER_OFFSET(0));

   TransferUsingPBOCmd::unbindPBO(GL_PIXEL_UNPACK_BUFFER_EXT);
}

//==============================================================================
// CLASS DrawWithPBOCmd
//==============================================================================

//------------------------------------------------------------------------------
//
DrawWithPBOCmd::DrawWithPBOCmd(GLenum extFmt, GLenum pixelType, int width,
                               int height, unsigned int alignment)
:
   DrawCmd(extFmt, pixelType, width, height, alignment),
   TransferUsingPBOCmd()
{}

//------------------------------------------------------------------------------
//
DrawWithPBOCmd::~DrawWithPBOCmd()
{}

//------------------------------------------------------------------------------
//
void DrawWithPBOCmd::init()
{
   DrawCmd::init();
   TransferUsingPBOCmd::initPBO(_pixelBufs.get(0), _imgSize, GL_PIXEL_UNPACK_BUFFER_EXT, GL_DYNAMIC_DRAW_ARB);
}

//------------------------------------------------------------------------------
//
void DrawWithPBOCmd::execute()
{
   TransferUsingPBOCmd::bindPBO(GL_PIXEL_UNPACK_BUFFER_EXT);

   // Use a different buffer on each execution so that the CPU can't cache the
   // data.
   static unsigned int srcIndex = 0;
   srcIndex++;
   copyToPBO( _pixelBufs.get(srcIndex), _imgSize);

   glDrawPixels(_width, _height, _extFmt, _pixelType, BUFFER_OFFSET(0));

   TransferUsingPBOCmd::unbindPBO(GL_PIXEL_UNPACK_BUFFER_EXT);
}

//==============================================================================
// CLASS DrawWithPBOConvertCmd
//==============================================================================

//------------------------------------------------------------------------------
//
DrawWithPBOConvertCmd::DrawWithPBOConvertCmd(GLenum extFmt, GLenum pixelType, int width,
                                             int height, unsigned int alignment)
:
   DrawWithPBOCmd(extFmt, pixelType, width, height, alignment)
{}

//------------------------------------------------------------------------------
//
DrawWithPBOConvertCmd::~DrawWithPBOConvertCmd()
{}

//------------------------------------------------------------------------------
//
bool DrawWithPBOConvertCmd::isSupported() const
{
    return( (_extFmt == GL_RGB || _extFmt == GL_RGBA) &&
             _pixelType == GL_UNSIGNED_SHORT );
}

//------------------------------------------------------------------------------
//
void DrawWithPBOConvertCmd::init()
{
   if (!isSupported()) {
      std::cerr << "[error] This test supports drawing from RGB or RGBA unsigned short only.\n";
      exit( EXIT_FAILURE );
   }

   DrawWithPBOCmd::init();
}

//------------------------------------------------------------------------------
//
void DrawWithPBOConvertCmd::execute()
{
   if (!isSupported()) {
      DrawWithPBOCmd::execute();
   }
   else {
      TransferUsingPBOCmd::bindPBO(GL_PIXEL_UNPACK_BUFFER_EXT);

      // Like copyToPBO() in DrawWithPBOCmd, but with a
      // conversion
      unsigned char *pPixelsPBO = static_cast<unsigned char *>(mapPBO(GL_PIXEL_UNPACK_BUFFER_EXT, GL_WRITE_ONLY));
      if (NULL == pPixelsPBO) {
         std::cerr << "[error] Couldn't map pixel buffer. Exiting\n";
         exit( EXIT_FAILURE );
      }

      // Use a different buffer on each execution so that the CPU can't cache the
      // data.
      static unsigned int srcIndex = 0;
      srcIndex++;

      if (reinterpret_cast<unsigned long long>(_pixelBufs.get(srcIndex)) % 16 != 0) {
         std::cerr << "[error] Source buffer must be aligned. Exiting\n";
         exit( EXIT_FAILURE );
      }

      cv_sse_RGBA16toRGBAHF(_pixelBufs.get(srcIndex), pPixelsPBO, _imgSize / (4*sizeof(unsigned short)));

      if (!unmapPBO(GL_PIXEL_UNPACK_BUFFER_EXT)) {
         std::cerr << "[error] Couldn't unmap pixel buffer. Exiting\n";
         exit( EXIT_FAILURE );
      }
      // End of the copyToPBO equivalent


      glDrawPixels(_width, _height, _extFmt, GL_HALF_FLOAT_ARB, BUFFER_OFFSET(0));

      TransferUsingPBOCmd::unbindPBO(GL_PIXEL_UNPACK_BUFFER_EXT);
   }
}

//==============================================================================
// CLASS DrawWithPBOConvertExpandCmd
//==============================================================================

//------------------------------------------------------------------------------
//
DrawWithPBOConvertExpandCmd::DrawWithPBOConvertExpandCmd(GLenum extFmt, GLenum pixelType, int width,
                                                         int height, unsigned int alignment)
:
   DrawWithPBOCmd(extFmt, pixelType, width, height, alignment),
   _convExtFmt(_extFmt == GL_RGB ? GL_RGBA : GL_BGRA)
{
}

//------------------------------------------------------------------------------
//
DrawWithPBOConvertExpandCmd::~DrawWithPBOConvertExpandCmd()
{
}

//------------------------------------------------------------------------------
//
bool DrawWithPBOConvertExpandCmd::isSupported() const
{
    return( (_extFmt == GL_RGB || _extFmt == GL_BGR) &&
            _pixelType == GL_UNSIGNED_SHORT );
}

//------------------------------------------------------------------------------
//
void DrawWithPBOConvertExpandCmd::init()
{
   if (!isSupported()) {
      std::cerr << "[error] This test supports drawing from RGB/BGR unsigned short only.\n";
      exit( EXIT_FAILURE );
   }

   DrawCmd::init();
   TransferUsingPBOCmd::initPBO(_pixelBufs.get(0), _width*_height*4*sizeof(unsigned short), GL_PIXEL_UNPACK_BUFFER_EXT, GL_DYNAMIC_DRAW_ARB);
}

//------------------------------------------------------------------------------
//
void DrawWithPBOConvertExpandCmd::execute()
{
   if (!isSupported()) {
      DrawWithPBOCmd::execute();
   }
   else {
      TransferUsingPBOCmd::bindPBO(GL_PIXEL_UNPACK_BUFFER_EXT);

      // Like copyToPBO() in DrawWithPBOCmd, but with a
      // conversion and expansion
      fillPBO();

      glDrawPixels(_width, _height, _convExtFmt, GL_HALF_FLOAT_ARB, BUFFER_OFFSET(0));

      TransferUsingPBOCmd::unbindPBO(GL_PIXEL_UNPACK_BUFFER_EXT);
   }
}

//------------------------------------------------------------------------------
//
void DrawWithPBOConvertExpandCmd::fillPBO()
{
   unsigned char *pPixelsPBO = static_cast<unsigned char *>(mapPBO(GL_PIXEL_UNPACK_BUFFER_EXT, GL_WRITE_ONLY));
   if (NULL == pPixelsPBO) {
      std::cerr << "[error] Couldn't map pixel buffer. Exiting\n";
      exit( EXIT_FAILURE );
   }

   // Use a different buffer on each execution so that the CPU can't cache the
   // data.
   static unsigned int srcIndex = 0;
   srcIndex++;

   if (reinterpret_cast<unsigned long long>(_pixelBufs.get(srcIndex)) % 16 != 0) {
      std::cerr << "[error] Source buffer must be aligned. Exiting\n";
      exit( EXIT_FAILURE );
   }

   cv_sse_RGB16toRGBAHF(_pixelBufs.get(srcIndex), pPixelsPBO, _width*_height);

   if (!unmapPBO(GL_PIXEL_UNPACK_BUFFER_EXT)) {
      std::cerr << "[error] Couldn't unmap pixel buffer. Exiting\n";
      exit( EXIT_FAILURE );
   }
}

//==============================================================================
// CLASS DrawWithPBOConvertExpandTexCmd
//==============================================================================

//------------------------------------------------------------------------------
//
DrawWithPBOConvertExpandTexCmd::DrawWithPBOConvertExpandTexCmd(GLenum intFmt, GLenum extFmt,
                                                               GLenum pixelType, int width,
                                                               int height, unsigned int alignment)
:
   DrawWithPBOConvertExpandCmd(extFmt, pixelType, width, height, alignment),
   _floatTexFP(0),
   _intFmt(intFmt)
{
}

//------------------------------------------------------------------------------
//
DrawWithPBOConvertExpandTexCmd::~DrawWithPBOConvertExpandTexCmd()
{
}

//------------------------------------------------------------------------------
//
void DrawWithPBOConvertExpandTexCmd::init()
{
   DrawWithPBOConvertExpandCmd::init();

   // Copied from TexLoadCmdBase::bindFloatTexFP()
   glGenProgramsNV(1, &_floatTexFP);

   const char FLOAT_TEX_PROG[] =
       "!!FP1.0\n"
       "TEX o[COLH], f[TEX0].xyxx, TEX0, RECT;\n"
       "END";
   glLoadProgramNV(GL_FRAGMENT_PROGRAM_NV, _floatTexFP, strlen(FLOAT_TEX_PROG),
                   (const GLubyte*)FLOAT_TEX_PROG);
}

//------------------------------------------------------------------------------
//
void DrawWithPBOConvertExpandTexCmd::execute()
{
   if (!isSupported()) {
      DrawWithPBOCmd::execute();
   }
   else {
      GLenum texTarget = GL_TEXTURE_RECTANGLE_NV;
      
      glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT | GL_TEXTURE_BIT |
                   GL_LIGHTING_BIT);

      GLuint tex = 0;
      glGenTextures(1, &tex);
      glBindTexture(texTarget, tex);

      // Initialize the texture environment
      glEnable(texTarget);
      glTexParameteri(texTarget, GL_TEXTURE_WRAP_S, GL_CLAMP);
      glTexParameteri(texTarget, GL_TEXTURE_WRAP_T, GL_CLAMP);
      glTexParameteri(texTarget, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
      glTexParameteri(texTarget, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
      glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);


      TransferUsingPBOCmd::bindPBO(GL_PIXEL_UNPACK_BUFFER_EXT);

      // Like copyToPBO() in DrawWithPBOCmd, but with a
      // conversion
      fillPBO();


      glTexImage2D(texTarget, 0,
                   _intFmt, _width, _height, 0,
                   _convExtFmt, GL_HALF_FLOAT_ARB, BUFFER_OFFSET(0));

      TransferUsingPBOCmd::unbindPBO(GL_PIXEL_UNPACK_BUFFER_EXT);


      glShadeModel(GL_FLAT);
      glDisable(GL_LIGHTING);
      glDisable(GL_DEPTH_TEST);
      glDisable(GL_CULL_FACE);
      glDisable(GL_POLYGON_SMOOTH);


      if (isNVFloatTexture(_intFmt)) {
          glBindProgramNV(GL_FRAGMENT_PROGRAM_NV, _floatTexFP);
          glEnable(GL_FRAGMENT_PROGRAM_NV);
      }


      // Actually do the drawing
      glBegin(GL_POLYGON);
         glTexCoord2f(0.0f, 0.0f);      glVertex2f(0, 0);
         glTexCoord2f(_width, 0.0f);    glVertex2f(_width, 0);
         glTexCoord2f(_width, _height); glVertex2f(_width, _height);
         glTexCoord2f(0.0f, _height);   glVertex2f(0, _height);
      glEnd();


      if (isNVFloatTexture(_intFmt)){
          glDisable(GL_FRAGMENT_PROGRAM_NV);
      }



      glDeleteTextures(1, &tex);

      glPopAttrib();
   }
}

//==============================================================================
// CLASS DrawWithPBOConvertExpandSubTexCmd
//==============================================================================

//------------------------------------------------------------------------------
//
DrawWithPBOConvertExpandSubTexCmd::DrawWithPBOConvertExpandSubTexCmd(GLenum intFmt, GLenum extFmt,
                                                                     GLenum pixelType, int width,
                                                                     int height, unsigned int alignment)
:
   DrawWithPBOConvertExpandTexCmd(intFmt, extFmt, pixelType, width, height, alignment),
   _tex(0)
{
   GLenum texTarget = GL_TEXTURE_RECTANGLE_NV;
      
   glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT | GL_TEXTURE_BIT |
                GL_LIGHTING_BIT);

   glGenTextures(1, &_tex);
   glBindTexture(texTarget, _tex);

   glTexImage2D(texTarget, 0,
                _intFmt, _width, _height, 0,
                _convExtFmt, GL_HALF_FLOAT_ARB, 0);

   glPopAttrib();
}

//------------------------------------------------------------------------------
//
DrawWithPBOConvertExpandSubTexCmd::~DrawWithPBOConvertExpandSubTexCmd()
{
   glDeleteTextures(1, &_tex);

   glPopAttrib();
}

//------------------------------------------------------------------------------
//
void DrawWithPBOConvertExpandSubTexCmd::execute()
{
   if (!isSupported()) {
      DrawWithPBOCmd::execute();
   }
   else {
      GLenum texTarget = GL_TEXTURE_RECTANGLE_NV;
      
      glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT | GL_TEXTURE_BIT |
                   GL_LIGHTING_BIT);

      glBindTexture(texTarget, _tex);

      // Initialize the texture environment
      glEnable(texTarget);
      glTexParameteri(texTarget, GL_TEXTURE_WRAP_S, GL_CLAMP);
      glTexParameteri(texTarget, GL_TEXTURE_WRAP_T, GL_CLAMP);
      glTexParameteri(texTarget, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
      glTexParameteri(texTarget, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
      glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);


      TransferUsingPBOCmd::bindPBO(GL_PIXEL_UNPACK_BUFFER_EXT);

      // Like copyToPBO() in DrawWithPBOCmd, but with a
      // conversion
      fillPBO();


      glTexSubImage2D(texTarget, 0,
                      0, 0, _width, _height,
                      _convExtFmt, GL_HALF_FLOAT_ARB, BUFFER_OFFSET(0));


      TransferUsingPBOCmd::unbindPBO(GL_PIXEL_UNPACK_BUFFER_EXT);


      glShadeModel(GL_FLAT);
      glDisable(GL_LIGHTING);
      glDisable(GL_DEPTH_TEST);
      glDisable(GL_CULL_FACE);
      glDisable(GL_POLYGON_SMOOTH);


      if (isNVFloatTexture(_intFmt)) {
          glBindProgramNV(GL_FRAGMENT_PROGRAM_NV, _floatTexFP);
          glEnable(GL_FRAGMENT_PROGRAM_NV);
      }


      // Actually do the drawing
      glBegin(GL_POLYGON);
         glTexCoord2f(0.0f, 0.0f);      glVertex2f(0, 0);
         glTexCoord2f(_width, 0.0f);    glVertex2f(_width, 0);
         glTexCoord2f(_width, _height); glVertex2f(_width, _height);
         glTexCoord2f(0.0f, _height);   glVertex2f(0, _height);
      glEnd();


      if (isNVFloatTexture(_intFmt)){
          glDisable(GL_FRAGMENT_PROGRAM_NV);
      }

      glPopAttrib();
   }
}

//==============================================================================
// CLASS TexLoadCmdBase
//==============================================================================

//------------------------------------------------------------------------------
//
const char FLOAT_TEX_PROG[] = 
  "!!FP1.0\n"		    
  "TEX o[COLH], f[TEX0].xyxx, TEX0, RECT;\n"
  "END"	;						

//------------------------------------------------------------------------------
//
TexLoadCmdBase::TexLoadCmdBase(
   GLenum intFmt,
   GLenum extFmt,
   GLenum pixelType,
   int width, int height,
   unsigned int alignment
) :
   _texTarget(), _intFmt(intFmt), _extFmt(extFmt), _width(width), _height(height),
   _pixelBufs( nbSrcBufs,
               (hasAlpha(extFmt) ? 4 : 3)*width*height*bytesPerChannel(pixelType),
               alignment ),
   _imgSize((hasAlpha(extFmt) ? 4 : 3)*width*height*bytesPerChannel(pixelType)),
   _pixelType(pixelType),
   _floatTexFP(0)
{
#ifdef GL_NV_texture_rectangle
  if (isNVFloatTexture(_intFmt))
    _texTarget = GL_TEXTURE_RECTANGLE_NV;
  else
#endif
    _texTarget = GL_TEXTURE_2D;
}

//------------------------------------------------------------------------------
//
TexLoadCmdBase::~TexLoadCmdBase()
{
}

//------------------------------------------------------------------------------
//
GLenum TexLoadCmdBase::getExtType()
{
   return _pixelType;
}

//------------------------------------------------------------------------------
//
GLenum TexLoadCmdBase::getExtFmt()
{
   return _extFmt;
}

//------------------------------------------------------------------------------
//
GLenum TexLoadCmdBase::getIntFmt()
{
   return _intFmt;
}

//------------------------------------------------------------------------------
//
GLenum TexLoadCmdBase::getTexTarget()
{
   return _texTarget;
}

//------------------------------------------------------------------------------
//
void TexLoadCmdBase::bindFloatTexFP()
{
#ifdef GFX_PB_ARB_USE_FP
   if (_floatTexFP == 0)
   {
      glGenProgramsNV(1, &_floatTexFP);

      glLoadProgramNV(GL_FRAGMENT_PROGRAM_NV, _floatTexFP, strlen(FLOAT_TEX_PROG),
                      (const GLubyte*)FLOAT_TEX_PROG);	
   }

   glBindProgramNV(GL_FRAGMENT_PROGRAM_NV, _floatTexFP);
   glEnable(GL_FRAGMENT_PROGRAM_NV);
#endif
}

//------------------------------------------------------------------------------
//
void TexLoadCmdBase::unbindFloatTexFP()
{
   glDisable(GL_FRAGMENT_PROGRAM_NV);
}

//==============================================================================
// CLASS TexLoadCmd
//==============================================================================

//------------------------------------------------------------------------------
//
TexLoadCmd::TexLoadCmd(
   GLenum intFmt,
   GLenum extFmt,
   GLenum pixelType,
   int width, int height,
   unsigned int alignment
) :
   TexLoadCmdBase(intFmt, extFmt, pixelType, width, height, alignment)
{
  for ( unsigned int i=0; i<_pixelBufs.size(); ++i ) {
    switch(pixelType){
    case GL_UNSIGNED_BYTE:
    default:
      fillImage(_extFmt, _pixelBufs.get(i), _width, _height, Color3ub(220, 20, 60));
      break;
    case GL_UNSIGNED_SHORT:
      fillImage(_extFmt, _pixelBufs.get(i), _width, _height, 
                Color3us(56320, 5120, 15360));
      break;
    case GL_HALF_FLOAT_ARB:
      fillImageHalf(_extFmt, _pixelBufs.get(i), _width, _height, 
                    Color3f(0.859f, 0.078f, 0.234f));
      break;
    case GL_UNSIGNED_INT:
      fillImage(_extFmt, _pixelBufs.get(i), _width, _height, 
                Color3ui(3690000000U, 335000000U, 1000000000U));
      break;
#ifdef GL_EXT_packed_pixels
    case GL_UNSIGNED_INT_10_10_10_2_EXT:
      fillImage(_extFmt, _pixelBufs.get(i), _width, _height, 
                Color4_10bit(880, 80, 240, 1023));
      break;
#endif
    case GL_FLOAT:
      fillImage(_extFmt, _pixelBufs.get(i), _width, _height, 
                Color3f(0.859f, 0.078f, 0.234f));
      break;
    }
  }
}

//------------------------------------------------------------------------------
//
TexLoadCmd::~TexLoadCmd()
{
}

//------------------------------------------------------------------------------
//
void TexLoadCmd::init()
{
   glTexParameteri(_texTarget, GL_TEXTURE_WRAP_S, GL_CLAMP);
   glTexParameteri(_texTarget, GL_TEXTURE_WRAP_T, GL_CLAMP);
   glTexParameteri(_texTarget, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameteri(_texTarget, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

   glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);

   // Reserve space in texture memory (don't load anything).
   int w2 = 1 << static_cast<int> ( ceil( log( static_cast<float>(_width)  ) /
                                          log ( 2.0F ) ) );
   int h2 = 1 << static_cast<int> ( ceil( log( static_cast<float>(_height) ) /
                                          log ( 2.0F ) ) );

   glTexImage2D(
       _texTarget, 0, _intFmt, w2, h2, 0, _extFmt,
       _pixelType, 0
   );
}

//------------------------------------------------------------------------------
//
void TexLoadCmd::execute()
{
   // Use a different buffer on each execution so that the CPU can't cache the
   // data.
   static unsigned int srcIndex = 0;
   srcIndex++;
   glTexSubImage2D(
      _texTarget, 0, 0, 0, _width, _height, _extFmt,
      _pixelType, _pixelBufs.get(srcIndex)
   );
}

//==============================================================================
//	CLASS TexLoadFromPBOCmd
//==============================================================================

//------------------------------------------------------------------------------
//
TexLoadFromPBOCmd::TexLoadFromPBOCmd(GLenum intFmt, GLenum extFmt, GLenum pixelType, int width,
                                     int height, unsigned int alignment)
:
   TexLoadCmd(intFmt, extFmt, pixelType, width, height, alignment),
   TransferUsingPBOCmd()
{}

//------------------------------------------------------------------------------
//
TexLoadFromPBOCmd::~TexLoadFromPBOCmd()
{}

//------------------------------------------------------------------------------
//
void TexLoadFromPBOCmd::init()
{
   TexLoadCmd::init();
   TransferUsingPBOCmd::initPBO(_pixelBufs.get(0), _imgSize, GL_PIXEL_UNPACK_BUFFER_EXT, GL_DYNAMIC_DRAW_ARB);
}

//------------------------------------------------------------------------------
//
void TexLoadFromPBOCmd::execute()
{
   TransferUsingPBOCmd::bindPBO(GL_PIXEL_UNPACK_BUFFER_EXT);

   // Need to map/unmap the pbo memory, otherwise the texture is not transfer
   // again
   copyToPBO(NULL, 0);

   glTexSubImage2D(
      _texTarget, 0, 0, 0, _width, _height, _extFmt,
      _pixelType, BUFFER_OFFSET(0)
   );

   TransferUsingPBOCmd::unbindPBO(GL_PIXEL_UNPACK_BUFFER_EXT);
}

//==============================================================================
//	CLASS TexLoadWithPBOCmd
//==============================================================================

//------------------------------------------------------------------------------
//
TexLoadWithPBOCmd::TexLoadWithPBOCmd(GLenum intFmt, GLenum extFmt, GLenum pixelType, int width,
                                     int height, unsigned int alignment)
:
   TexLoadCmd(intFmt, extFmt, pixelType, width, height, alignment),
   TransferUsingPBOCmd()
{}

//------------------------------------------------------------------------------
//
TexLoadWithPBOCmd::~TexLoadWithPBOCmd()
{}

//------------------------------------------------------------------------------
//
void TexLoadWithPBOCmd::init()
{
   TexLoadCmd::init();
   TransferUsingPBOCmd::initPBO(_pixelBufs.get(0), _imgSize, GL_PIXEL_UNPACK_BUFFER_EXT, GL_DYNAMIC_DRAW_ARB);
}

//------------------------------------------------------------------------------
//
void TexLoadWithPBOCmd::execute()
{
   TransferUsingPBOCmd::bindPBO(GL_PIXEL_UNPACK_BUFFER_EXT);

   // Use a different buffer on each execution so that the CPU can't cache the
   // data.
   static unsigned int srcIndex = 0;
   srcIndex++;
   copyToPBO( _pixelBufs.get(srcIndex), _imgSize );

   glTexSubImage2D(
      _texTarget, 0, 0, 0, _width, _height, _extFmt,
      _pixelType, BUFFER_OFFSET(0)
   );

   TransferUsingPBOCmd::unbindPBO(GL_PIXEL_UNPACK_BUFFER_EXT);
}

//==============================================================================
// CLASS TexLoadUseCmd
//==============================================================================

//------------------------------------------------------------------------------
//
TexLoadUseCmd::TexLoadUseCmd(
   GLenum intFmt,
   GLenum extFmt,
   GLenum pixelType,
   int width,
   int height,
   unsigned int alignment
) :
   TexLoadCmd(intFmt, extFmt, pixelType, width, height, alignment),
   _x(1), _y(1)
{
}

//------------------------------------------------------------------------------
//
TexLoadUseCmd::~TexLoadUseCmd()
{
}

//------------------------------------------------------------------------------
//
void TexLoadUseCmd::init()
{
   TexLoadCmd::init();

   glEnable(_texTarget);

   // Don't even have to actually draw anything --- scissor it out.
   glEnable(GL_SCISSOR_TEST);
   glScissor(0, 0, 0, 0);
}

//------------------------------------------------------------------------------
//
void TexLoadUseCmd::execute()
{
   TexLoadCmd::execute();       // Do the glTexSubImage()

   if (isNVFloatTexture(_intFmt))
      TexLoadCmdBase::bindFloatTexFP();

   // Actually use the texture: texturing a single point is enough to
   // pull in the texture.
   glBegin(GL_POINTS);
   glVertex2i(_x, _y);
   glEnd();

   if (isNVFloatTexture(_intFmt))
      TexLoadCmdBase::unbindFloatTexFP();
}

//==============================================================================
//	CLASS TexLoadUseFromPBOCmd
//==============================================================================

//------------------------------------------------------------------------------
//
TexLoadUseFromPBOCmd::TexLoadUseFromPBOCmd(GLenum intFmt, GLenum extFmt, GLenum pixelType, int width,
                                     int height, unsigned int alignment)
:
   TexLoadUseCmd(intFmt, extFmt, pixelType, width, height, alignment),
   TransferUsingPBOCmd()
{}

//------------------------------------------------------------------------------
//
TexLoadUseFromPBOCmd::~TexLoadUseFromPBOCmd()
{}

//------------------------------------------------------------------------------
//
void TexLoadUseFromPBOCmd::init()
{
   TexLoadUseCmd::init();
   TransferUsingPBOCmd::initPBO(_pixelBufs.get(0), _imgSize, GL_PIXEL_UNPACK_BUFFER_EXT, GL_DYNAMIC_DRAW_ARB);
}

//------------------------------------------------------------------------------
//
void TexLoadUseFromPBOCmd::execute()
{
   TransferUsingPBOCmd::bindPBO(GL_PIXEL_UNPACK_BUFFER_EXT);

   // Need to map/unmap the pbo memory, otherwise the texture is not transfer
   // again
   copyToPBO(NULL, 0);

   glTexSubImage2D(
      _texTarget, 0, 0, 0, _width, _height, _extFmt,
      _pixelType, BUFFER_OFFSET(0)
   );

   TransferUsingPBOCmd::unbindPBO(GL_PIXEL_UNPACK_BUFFER_EXT);

   if (isNVFloatTexture(_intFmt))
      TexLoadCmdBase::bindFloatTexFP();

   // Actually use the texture: texturing a single point is enough to
   // pull in the texture.
   glBegin(GL_POINTS);
   glVertex2i(_x, _y);
   glEnd();

   if (isNVFloatTexture(_intFmt))
      TexLoadCmdBase::unbindFloatTexFP();
}

//==============================================================================
//	CLASS TexLoadUseWithPBOCmd
//==============================================================================

//------------------------------------------------------------------------------
//
TexLoadUseWithPBOCmd::TexLoadUseWithPBOCmd(GLenum intFmt, GLenum extFmt, GLenum pixelType, int width,
                                     int height, unsigned int alignment)
:
   TexLoadUseCmd(intFmt, extFmt, pixelType, width, height, alignment),
   TransferUsingPBOCmd()
{}

//------------------------------------------------------------------------------
//
TexLoadUseWithPBOCmd::~TexLoadUseWithPBOCmd()
{}

//------------------------------------------------------------------------------
//
void TexLoadUseWithPBOCmd::init()
{
   TexLoadUseCmd::init();
   TransferUsingPBOCmd::initPBO(NULL, _imgSize, GL_PIXEL_UNPACK_BUFFER_EXT, GL_DYNAMIC_DRAW_ARB);
}

//------------------------------------------------------------------------------
//
void TexLoadUseWithPBOCmd::execute()
{
   TransferUsingPBOCmd::bindPBO(GL_PIXEL_UNPACK_BUFFER_EXT);

   // Use a different buffer on each execution so that the CPU can't cache the
   // data.
   static unsigned int srcIndex = 0;
   srcIndex++;

   copyToPBO( _pixelBufs.get(srcIndex), _imgSize );

   glTexSubImage2D(
      _texTarget, 0, 0, 0, _width, _height, _extFmt,
      _pixelType, BUFFER_OFFSET(0)
   );

   TransferUsingPBOCmd::unbindPBO(GL_PIXEL_UNPACK_BUFFER_EXT);

   if (isNVFloatTexture(_intFmt))
      TexLoadCmdBase::bindFloatTexFP();

   // Actually use the texture: texturing a single point is enough to
   // pull in the texture.
   glBegin(GL_POINTS);
   glVertex2i(_x, _y);
   glEnd();

   if (isNVFloatTexture(_intFmt))
      TexLoadCmdBase::unbindFloatTexFP();
}

//==============================================================================
//	CLASS TexLoadUseWithPBOConvertCmd
//==============================================================================

//------------------------------------------------------------------------------
//
TexLoadUseWithPBOConvertCmd::TexLoadUseWithPBOConvertCmd(GLenum intFmt, GLenum extFmt, GLenum pixelType, int width,
                                     int height, unsigned int alignment)
:
   TexLoadUseWithPBOCmd(intFmt, extFmt, pixelType, width, height, alignment)
{}

//------------------------------------------------------------------------------
//
TexLoadUseWithPBOConvertCmd::~TexLoadUseWithPBOConvertCmd()
{}

//------------------------------------------------------------------------------
//
bool TexLoadUseWithPBOConvertCmd::isSupported() const
{
    return( _pixelType == GL_UNSIGNED_SHORT );
}

//------------------------------------------------------------------------------
//
void TexLoadUseWithPBOConvertCmd::init()
{
   if (!isSupported()) {
      std::cerr << "[error] This test supports loading from unsigned short only.\n";
      exit( EXIT_FAILURE );
   }

   TexLoadUseWithPBOCmd::init();
}

//------------------------------------------------------------------------------
//
void TexLoadUseWithPBOConvertCmd::execute()
{
   if ( !isSupported() ) {
      TexLoadUseWithPBOCmd::execute();
   }
   else {
      TransferUsingPBOCmd::bindPBO(GL_PIXEL_UNPACK_BUFFER_EXT);


      // Like copyToPBO() in TexLoadUseWithPBOCmd, but with a
      // conversion
      unsigned char *pPixelsPBO = static_cast<unsigned char *>(mapPBO(GL_PIXEL_UNPACK_BUFFER_EXT, GL_WRITE_ONLY));
      if (NULL == pPixelsPBO) {
         std::cerr << "[error] Couldn't map pixel buffer. Exiting\n";
         exit( EXIT_FAILURE );
      }

      // Use a different buffer on each execution so that the CPU can't cache the
      // data.
      static unsigned int srcIndex = 0;
      srcIndex++;
      cv_sse_RGBA16toRGBAHF(_pixelBufs.get(srcIndex), pPixelsPBO, _imgSize / (4*sizeof(unsigned short)));

      if (!unmapPBO(GL_PIXEL_UNPACK_BUFFER_EXT)) {
         std::cerr << "[error] Couldn't unmap pixel buffer. Exiting\n";
         exit( EXIT_FAILURE );
      }
      // End of the copyToPBO equivalent


      glTexSubImage2D(
         _texTarget, 0, 0, 0, _width, _height, _extFmt,
         GL_HALF_FLOAT_ARB, BUFFER_OFFSET(0)
      );

      TransferUsingPBOCmd::unbindPBO(GL_PIXEL_UNPACK_BUFFER_EXT);

      if (isNVFloatTexture(_intFmt))
         TexLoadCmdBase::bindFloatTexFP();

      // Actually use the texture: texturing a single point is enough to
      // pull in the texture.
      glBegin(GL_POINTS);
      glVertex2i(_x, _y);
      glEnd();

      if (isNVFloatTexture(_intFmt))
         TexLoadCmdBase::unbindFloatTexFP();
   }
}

//==============================================================================
//	CLASS TexLoadUseWithPBOConvertExpandCmd
//==============================================================================

//------------------------------------------------------------------------------
//
TexLoadUseWithPBOConvertExpandCmd::TexLoadUseWithPBOConvertExpandCmd(GLenum intFmt, GLenum extFmt, GLenum pixelType, int width,
                                     int height, unsigned int alignment)
:
   TexLoadUseWithPBOCmd(intFmt, extFmt, pixelType, width, height, alignment),
   _nbPixels(width*height),
   _convExtFmt(_extFmt == GL_RGB ? GL_RGBA : GL_BGRA)
{}

//------------------------------------------------------------------------------
//
TexLoadUseWithPBOConvertExpandCmd::~TexLoadUseWithPBOConvertExpandCmd()
{}

//------------------------------------------------------------------------------
//
bool TexLoadUseWithPBOConvertExpandCmd::isSupported() const
{
    return( (_intFmt == GL_FLOAT_RGBA16_NV || _intFmt == GL_RGBA16F_ARB) &&
            (_extFmt == GL_RGB || _extFmt == GL_BGR) &&
            _pixelType == GL_UNSIGNED_SHORT );
}

//------------------------------------------------------------------------------
//
void TexLoadUseWithPBOConvertExpandCmd::init()
{
   if (!isSupported()) {
      std::cerr << "[error] This test supports loading from RGB/BGR unsigned short to RGBA half-float only.\n";
      exit( EXIT_FAILURE );
   }

   if ( !isSupported() ) {
       TexLoadUseWithPBOCmd::init();
   }
   else {
       TexLoadUseCmd::init();
       TransferUsingPBOCmd::initPBO(NULL, _nbPixels*4*sizeof(unsigned short), GL_PIXEL_UNPACK_BUFFER_EXT, GL_DYNAMIC_DRAW_ARB);
   }
}

//------------------------------------------------------------------------------
//
void TexLoadUseWithPBOConvertExpandCmd::execute()
{
    if ( !isSupported() ) {
       TexLoadUseWithPBOCmd::execute();
    }
    else {
       TransferUsingPBOCmd::bindPBO(GL_PIXEL_UNPACK_BUFFER_EXT);


       // Like copyToPBO() in TexLoadUseWithPBOCmd, but with a
       // conversion
       unsigned char *pPixelsPBO = static_cast<unsigned char *>(mapPBO(GL_PIXEL_UNPACK_BUFFER_EXT, GL_WRITE_ONLY));
       if (NULL == pPixelsPBO) {
          std::cerr << "[error] Couldn't map pixel buffer. Exiting\n";
          exit( EXIT_FAILURE );
       }

       // Use a different buffer on each execution so that the CPU can't cache the
       // data.
       static unsigned int srcIndex = 0;
       srcIndex++;
       cv_sse_RGB16toRGBAHF(_pixelBufs.get(srcIndex), pPixelsPBO, _nbPixels);

       if (!unmapPBO(GL_PIXEL_UNPACK_BUFFER_EXT)) {
          std::cerr << "[error] Couldn't unmap pixel buffer. Exiting\n";
          exit( EXIT_FAILURE );
       }
       // End of the copyToPBO equivalent


       glTexSubImage2D(
         _texTarget, 0, 0, 0, _width, _height, _convExtFmt,
         GL_HALF_FLOAT_ARB, BUFFER_OFFSET(0)
       );

       TransferUsingPBOCmd::unbindPBO(GL_PIXEL_UNPACK_BUFFER_EXT);

       if (isNVFloatTexture(_intFmt))
          TexLoadCmdBase::bindFloatTexFP();

       // Actually use the texture: texturing a single point is enough to
       // pull in the texture.
       glBegin(GL_POINTS);
       glVertex2i(_x, _y);
       glEnd();

       if (isNVFloatTexture(_intFmt))
          TexLoadCmdBase::unbindFloatTexFP();
    }
}

//==============================================================================
// CLASS FillCmd
//==============================================================================

//------------------------------------------------------------------------------
//
FillCmd::FillCmd(bool pixelAlpha, int width, int height) :
   TransferCmd(), _width(width), _height(height), _alpha(pixelAlpha)
{
}

//------------------------------------------------------------------------------
//
FillCmd::~FillCmd()
{
}

//------------------------------------------------------------------------------
//
void FillCmd::init()
{
   if (_alpha) {
      glColor4ub(220, 20, 60, 255);
   }
   else {
      glColor3ub(220, 20, 60);
   }
}

//------------------------------------------------------------------------------
//
void FillCmd::execute()
{
   glRecti(0, 0, _width, _height);
}

//------------------------------------------------------------------------------
//
GLenum FillCmd::getExtFmt()
{
   if (_alpha) {
      return GL_RGBA;
   }
   else
   {
      return GL_RGB;
   }
}

//==============================================================================
// CLASS TexFillCmd
//==============================================================================

//------------------------------------------------------------------------------
//
TexFillCmd::TexFillCmd(GLenum intFmt, GLenum extFmt, GLenum pixelType,
                       int width, int height, unsigned int alignment) :
   TransferCmd(),
   _texTarget(), _intFmt(intFmt), _extFmt(extFmt),
   _width(width), _height(height),
   _pixelBufs( nbSrcBufs,
               (hasAlpha(extFmt) ? 4 : 3)*width*height*bytesPerChannel(pixelType), 
               alignment),
   _pixelType(pixelType)
{
#ifdef GL_NV_texture_rectangle
  if (isFloatTexture(_intFmt))
    _texTarget = GL_TEXTURE_RECTANGLE_NV;
  else
#endif
    _texTarget = GL_TEXTURE_2D;

  for ( unsigned int i=0; i<_pixelBufs.size(); ++i ) {
    switch(pixelType){
    case GL_UNSIGNED_BYTE:
    default:
      fillImage(_extFmt, _pixelBufs.get(i), _width, _height, Color3ub(220, 20, 60));
      break;
    case GL_UNSIGNED_SHORT:
      fillImage(_extFmt, _pixelBufs.get(i), _width, _height, 
                Color3us(56320, 5120, 15360));
      break;
    case GL_HALF_FLOAT_ARB:
      fillImageHalf(_extFmt, _pixelBufs.get(i), _width, _height, 
                    Color3f(0.859f, 0.078f, 0.234f));
      break;
    case GL_UNSIGNED_INT:
      fillImage(_extFmt, _pixelBufs.get(i), _width, _height, 
                Color3ui(3690000000U, 335000000U, 1000000000U));
      break;
#ifdef GL_EXT_packed_pixels
    case GL_UNSIGNED_INT_10_10_10_2_EXT:
      fillImage(_extFmt, _pixelBufs.get(i), _width, _height, 
                Color4_10bit(880, 80, 240, 1023));
      break;
#endif
    case GL_FLOAT:
      fillImage(_extFmt, _pixelBufs.get(i), _width, _height, 
                Color3f(0.859f, 0.078f, 0.234f));
      break;
    }
  }
}

//------------------------------------------------------------------------------
//
TexFillCmd::~TexFillCmd()
{
}

//------------------------------------------------------------------------------
//
void TexFillCmd::init()
{
   // Set up wrapping mode, texture environment.
   glTexParameteri(_texTarget, GL_TEXTURE_WRAP_S, GL_CLAMP);
   glTexParameteri(_texTarget, GL_TEXTURE_WRAP_T, GL_CLAMP);
   glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);

   // Set up filtering
   glTexParameteri(_texTarget, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameteri(_texTarget, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

   // Reserve space in texture memory (don't load anything).
   int w2 = 1 << static_cast<int> ( ceil( log( static_cast<float>(_width)  ) /
                                          log ( 2.0F ) ) );
   int h2 = 1 << static_cast<int> ( ceil( log( static_cast<float>(_height) ) /
                                          log ( 2.0F ) ) );

   _x0 = static_cast<float>(_width)  / static_cast<float>(w2);
   _y0 = static_cast<float>(_height) / static_cast<float>(h2);

   glTexImage2D(
      _texTarget, 0, _intFmt, w2, h2, 0, _extFmt,
      _pixelType, 0
   );

   // Load up texture.
   glTexSubImage2D(
      _texTarget, 0, 0, 0, width, height, _extFmt,
      _pixelType, _pixelBufs.get(0)
   );

   // Set up texturing.
   glEnable(_texTarget);
}

//------------------------------------------------------------------------------
//
void TexFillCmd::execute()
{
#ifdef GL_NV_texture_rectangle
   if (_texTarget == GL_TEXTURE_RECTANGLE_NV)
   {
      glBegin(GL_POLYGON);
      glTexCoord2f(0.0f, 0.0f);       glVertex2i(0, 0);
      glTexCoord2f(_width,  0.0f);    glVertex2i(_width, 0);
      glTexCoord2d(_width,  _height); glVertex2i(_width, _height);
      glTexCoord2d(0.0f, _height);    glVertex2i(0, _height);
      glEnd();
   }
   else
#endif
   {
      glBegin(GL_POLYGON);
      glTexCoord2f(0.0f, 0.0f); glVertex2i(0, 0);
      glTexCoord2f(_x0,  0.0f); glVertex2i(_width, 0);
      glTexCoord2d(_x0,  _y0);  glVertex2i(_width, _height);
      glTexCoord2d(0.0f, _y0);  glVertex2i(0, _height);
      glEnd();
   }
}

//------------------------------------------------------------------------------
//
GLenum TexFillCmd::getIntFmt()
{
   return _intFmt;
};

//------------------------------------------------------------------------------
//
GLenum TexFillCmd::getExtFmt()
{
   return _extFmt;
};

//------------------------------------------------------------------------------
//
GLenum TexFillCmd::getExtType()
{
   return _pixelType;
};

//------------------------------------------------------------------------------
//
GLenum TexFillCmd::getTexTarget()
{
   return _texTarget;
};

//==============================================================================
// CLASS ReadCmd
//==============================================================================

//------------------------------------------------------------------------------
//
ReadCmd::ReadCmd(GLenum extFmt, GLenum pixelType,
                 int width, int height, unsigned int alignment) :
   TransferCmd(), _extFmt(extFmt), _width(width), _height(height),
   _pixelBufs( nbDestBufs,
               (hasAlpha(extFmt) ? 4 : 3)*width*height*bytesPerChannel(pixelType), 
               alignment),
   _pixelType(pixelType)
{
}

//------------------------------------------------------------------------------
//
ReadCmd::~ReadCmd()
{
}

//------------------------------------------------------------------------------
//
void ReadCmd::init()
{
}

//------------------------------------------------------------------------------
//
void ReadCmd::execute()
{
   // Use a different buffer on each execution so that the CPU can't cache the
   // da
   static unsigned int destIndex = 0;
   destIndex++;
   glReadPixels(
      0, 0, _width, _height, _extFmt, _pixelType, _pixelBufs.get(destIndex)
   );
}

//------------------------------------------------------------------------------
//
GLenum ReadCmd::getExtType()
{
   return _pixelType;
}

//------------------------------------------------------------------------------
//
GLenum ReadCmd::getExtFmt()
{
   return _extFmt;
}

//==============================================================================
//	FREE FUNCTIONS
//==============================================================================

//------------------------------------------------------------------------------
//
void exitFunction()
{
   // Destroy window.
   glutDestroyWindow(windowId);

   // Destroy command object
   delete gfxCmd;

   // Close internal services.
   OsalInit::performFinalize();
   ProfilingInit::performFinalize();

   exit(0);
}

//------------------------------------------------------------------------------
//
void displayFunction()
{
   // Perform an initial refresh without drawing anything (required on MAC for
   // the window to appear)
   static bool firstPass = true;
   if ( firstPass ) {
      firstPass = false;
      // Mark for redraw.
      glutPostRedisplay();
      return;
   }

   HwIntervalTimer hwTimer;
   PrProcessTimer  processTimer;
   TodTimer        todTimer;
   HwProfiler	   mapProf("glMapBuffer");
   pMapProf = &mapProf;
   
   ArrayPtr<double> gfxTimes(new double[nbIterations]);

   // Initialize command gfx states.
   (*gfxCmd)->init();

   // Perform enough iterations so that each buffer which will be used in the
   // test is touched at least once (for PBOs as well as memory buffers, there
   // is a hit on first use)
   int iter = std::max( nbPBOs, std::max( nbSrcBufs, nbDestBufs ) );
   while (iter--) {
      glRasterPos2i(0, 0);
      (*gfxCmd)->execute();
      if ( glFinishBetweenIterations || iter==0 ) {
         glFinish();
      }
   }

   std::cout << mapProf;
   mapProf.reset();

   processTimer.start();
   todTimer.start();

   iter = nbIterations;
   while (iter--) {
      glRasterPos2i(0, 0);
      
      hwTimer.start();

      (*gfxCmd)->execute();
      if ( glFinishBetweenIterations || iter==0 ) {
         glFinish();
      }
      
      hwTimer.stop();

      gfxTimes[iter] = hwTimer.timeInterval();
   }
   
   todTimer.stop();
   processTimer.stop();

   // Process result.
   if ((*gfxCmd)->getUserData() != "")
      std::cout << (*gfxCmd)->getUserDataTitle() << ": " << (*gfxCmd)->getUserData() << std::endl;

   if ((*gfxCmd)->getExtType() != 0)
      std::cout << "Ext. Type: " << getPixelTypeStr((*gfxCmd)->getExtType()) << std::endl;

   if ((*gfxCmd)->getExtFmt() != 0)
      std::cout << "Ext. Format: " << getPixelFmtStr((*gfxCmd)->getExtFmt()) << std::endl;

   if ((*gfxCmd)->getIntFmt() != 0)
      std::cout << "Int. Format: " << getPixelFmtStr((*gfxCmd)->getIntFmt()) << std::endl;

   if ((*gfxCmd)->getTexTarget() != 0)
      std::cout << "Tex. Target: " << getTexTargetStr((*gfxCmd)->getTexTarget()) << std::endl;

   std::cout << "RT Clock Time: " << todTimer.getSeconds() << "s"
             << std::endl;

   double realTime = processTimer.realTime();
   double cpuTime  = processTimer.cpuTime();
   double totTime  = (cpuTime>realTime) ? cpuTime : realTime;
   
   double kernelPercent = 100.0 * processTimer.cpuKernelTime()/totTime;
   double userPercent = 100.0 * processTimer.cpuUserTime()/totTime;

   std::cout << "System Clock Tick Time: " << realTime << "s" << std::endl;
   std::cout << "CPU Time Used: " << cpuTime << std::endl;
   std::ios::fmtflags oldFlags = std::cout.setf(std::ios::fixed);
   long oldPrec  = std::cout.precision( 2 );
   std::cout << "% Kernel: " << kernelPercent << std::endl;
   std::cout << "% User: " << userPercent << std::endl;
   std::cout.flags( oldFlags );
   std::cout.precision( oldPrec );

   double sum1 = 0;
   for ( int i=0;i<nbIterations;++i ) {
      sum1 += gfxTimes[i];
   }

   double avg = sum1/double(nbIterations);

   std::cout << "Avg: " << avg << std::endl;

   double sum2 = 0;
   for ( int j=0;j<nbIterations;++j ) {
      double tmp = gfxTimes[j]-avg;
      sum2 += tmp*tmp;
   }
   
   double var = sum2/double(nbIterations-1);

   std::cout << "Std Dev: " << sqrt(var) << std::endl;

   int bytesPerPix = 0;
   if ((*gfxCmd)->getExtFmt() != 0)
   {
      bytesPerPix = bytesPerChannel((*gfxCmd)->getExtType())
         * (hasAlpha((*gfxCmd)->getExtFmt()) ? 4 : 3);
   }
   else
   {
      bytesPerPix = bytesPerPixels((*gfxCmd)->getIntFmt());
   }

   double mpix_sec = width*height/ ( avg * 1000000.0 );
   double mbytes_sec = mpix_sec * bytesPerPix;
   oldFlags = std::cout.setf( std::ios::fixed );
   oldPrec  = std::cout.precision( 2 );
   std::cout << "MBytes/Sec: " << mbytes_sec << std::endl;
   std::cout << "MPix/Sec: " << mpix_sec << std::endl;
   std::cout.flags( oldFlags );
   std::cout.precision( oldPrec );

   exitFunction();
}

//------------------------------------------------------------------------------
//
void reshapeFunction(int w, int h)
{
   static bool firstPass = true;

   if ( firstPass ) {
      // Initialize OpenGL state.
      initState();
      initDraw(width, height, zoom);
      firstPass = false;
      return;
   }
   
   glViewport(0, 0, w, h); 
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity(); 
   glOrtho(0, w, 0, h, -1, 1);
}

}

//==============================================================================
// PUBLIC FUNCTIONS
//==============================================================================

//------------------------------------------------------------------------------
//
int main( int argc, char* argv[] )
{
   // Initialize GL utilities.
   glutInit(&argc, argv);

   // Initialize internal services.
   OsalInit::performInitialize();
   ProfilingInit::performInitialize();

   GLenum pixelType = GL_UNSIGNED_BYTE;
   unsigned int alignment = 8;
   int userData = 0;
   
   int cmdOption = 0;
   int c = 0;
   while ((c = getopt(argc, argv,"c,d,f,h,j,k,l,m,n,r,t,w,b,o,v,s,q,A,B,C,D,E,F,G,H,x:,y:,e:,i:,p:,z:,a:,u:,g,S")) != -1) {
      switch (c) {
         case 'c':
         case 'd':
         case 'f':
         case 'h':
         case 'j':
         case 'k':
         case 'l':
         case 'm':
         case 'n':
         case 'r':
         case 't':
         case 'w':
         case 'b':
         case 's':
         case 'q':
         case 'o':
         case 'A':
         case 'B':
         case 'C':
         case 'D':
         case 'E':
         case 'F':
         case 'G':
         case 'H': {
            cmdOption = c;
         } break;

         case 'x': {
            width = atoi(optarg);
         } break;

         case 'y': {
            height = atoi(optarg);
         } break;

         case 'z': {
            zoom = atof(optarg);
         } break;

         case 'a': {
            alignment = atoi(optarg);

            if (alignment < 1 || alignment > 8192) {
               std::cerr << "[error] illegal byte alignment value "
                         << alignment << std::endl;
               return 1;
            }
         } break;

         case 'u': {
             userData = atoi(optarg);
         } break;

         case 'e': {
            extFormat = getPixelFmt(optarg);

            if (extFormat == 0) {
               std::cerr << "[error] illegal external format string "
                         << optarg << std::endl;
               return 1;
            }
         } break;

         case 'i': {
            intFormat = getPixelFmt(optarg);

            if (intFormat == 0) {
               std::cerr << "[error] illegal internal format string "
                         << optarg << std::endl;
               return 1;
            }
         } break;

         case 'p': {
            pixelType = getPixelType(optarg);

            if (pixelType == 0) {
               std::cerr << "[error] Pixel type " << optarg
                  << " not supported" << std::endl;
               return 1;
            }
         } break;

         case 'g': {
            glFinishBetweenIterations = true;
         } break;

         case 'S': {
            nbPBOs = 1;
         } break;

         default: {
            printUsage(argv[0]);
            return 1;
         }
      }
   }

   nbIterations = atoi(argv[argc-1]);

   const int MAX_NB_ITERATIONS = 1000000; // Sanity check
   if (nbIterations < 0 || nbIterations > MAX_NB_ITERATIONS) {
      std::cerr << "[error] illegal nb of iterations" << std::endl;
      return 1;
   }

   // Initialize GL utility display and window.
   if(cmdOption == 's') {
      glutInitDisplayMode(GLUT_DOUBLE|GLUT_RGBA|GLUT_DEPTH|GLUT_MULTISAMPLE);
      // TODO: setting GL_MULTISAMPLE_FILTER_HINT_NV to GL_FASTEST crashes
      //       on MacOSX in Cocoa. Disabled for now.
#ifndef OPENGL_FRAMEWORK
      glHint( GL_MULTISAMPLE_FILTER_HINT_NV, GL_FASTEST );
#endif
   } else if(cmdOption == 'q') {
      glutInitDisplayMode(GLUT_DOUBLE|GLUT_RGBA|GLUT_DEPTH|GLUT_MULTISAMPLE);   
      // TODO: setting GL_MULTISAMPLE_FILTER_HINT_NV to GL_FASTEST crashes
      //       on MacOSX in Cocoa. Disabled for now.
#ifndef OPENGL_FRAMEWORK
      glHint( GL_MULTISAMPLE_FILTER_HINT_NV, GL_NICEST );
#endif
   } else
      glutInitDisplayMode(GLUT_DOUBLE|GLUT_RGBA|GLUT_DEPTH);
   glutInitWindowPosition(10, 10);
   glutInitWindowSize(width, height);

   // Create window.
   windowId = glutCreateWindow("gfxProfile");

   // Make the window current before doing anything else, as we need a
   // current OpenGL context.
   glutSetWindow(windowId);

   if( GLeeInit() != GL_TRUE )
   {
      std::cerr << "[error] Failed to init GLEE: " << GLeeGetErrorString() << std::endl;
      exit( EXIT_FAILURE );
   }

   if(pixelType == GL_UNSIGNED_INT_10_10_10_2_EXT){
     if(!GLEE_EXT_packed_pixels){
       std::cerr << "[error] Extension GL_EXT_packed_pixels not found."
		 << std::endl;
       return 0;
     }
   }

   // check for pbo
   if( (cmdOption == 'b' || 
        cmdOption == 'd' || 
        cmdOption == 'j' || 
        cmdOption == 'k' || 
        cmdOption == 'm' ||  
        cmdOption == 'n' ||
        cmdOption == 'o' ||
        cmdOption == 'A' ||
        cmdOption == 'B' ||
        cmdOption == 'C' ||
        cmdOption == 'D' ||
        cmdOption == 'E' ||
        cmdOption == 'G' ||
        cmdOption == 'H')
       && !GLEE_ARB_pixel_buffer_object ) {
      std::cerr << "[error] Extension GL_ARB_pixel_buffer_object not found."
                << std::endl;
      return 0;
   }


   // Create command to be benchmarked.
   gfxCmd = new ObjPtr<TransferCmd>(
      cmdFactory(cmdOption, intFormat, extFormat, pixelType, width, height, 
		 alignment, userData)
   );
   
   if ((*gfxCmd).get() == 0) {
      printUsage(argv[0]);
      return 1;
   }

   // Set glut function callbacks.
   glutDisplayFunc(displayFunction);
   glutReshapeFunc(reshapeFunction);

   // Draw in loop.
   glutMainLoop();              	

   return 0;
}
